#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
""" An alias import module for the `template` module.

This exists because `template` is a keyword in Enaml, and the module
may need to be imported by Enaml code.

"""
from .template import *
